from string import *
from time import *
import sys

TRUE=1
FALSE=0

# This is our tree data-type. It can be used to read data of the format
#  Key: value
#  Key2: value2
#      Key2Subkey1: value3
#  Key3: value4
# It emulates a dictionary of pairs of values and sub-trees (Or None
# if no subtree exists).
class GeneralTree:
 def __init__(self):
  self.dict = {}
  self.current = None
 def read(self, str):
  if str == "":
   return
  if str[0] in whitespace:
   if self.current == None:
    str = lstrip(str)
    if str == "":
     return
   else:
    if self.dict[self.current][1] == None:
     self.dict[self.current][1] = GeneralTree()
    self.dict[self.current][1].read(str[1:])
    return
  items = split(str, ":", 1);
  if len(items) == 2:
   self.current = rstrip(items[0])
   value = strip(items[1])
   if not self.dict.has_key(self.current):
    self.dict[self.current] = [value, None]
  else:
   if str[0] == "?":
     # Unknown file of CVS when CVSROOT is not local
     print >> sys.stderr, join(("Warning : what about '", str, "'"),"")
   else:
     self.dict[self.current][0] = self.dict[self.current][0] + "\n" + items[0]
 def keys(self):
  return self.dict.keys()
 def has_key(self, key):
  return self.dict.has_key(key)
 def __len__(self):
  return len(self.dict)
 def __getitem__(self, key):
  return self.dict[key]
 def __setitem__(self, key, value):
  self.dict[key] = value
 def __delitem__(self, key):
  del self.dict[key]
 def __repr__(self):
  return repr(self.dict)

class PragmaticTree(GeneralTree):
 # Features of the cvs log output
 descriptionKey = "Description"
 revisionKey = "Revision"
 revisionsKey = "Revisions"
 dateKey = "date"

 revisionSep = "----------------------------"
 fileSep = "============================================================================="
 fieldSeparator = ";"
 branchesPrefix = "branches:"
 timeFormat = "%Y/%m/%d %H:%M:%S"

 HEADER            = 0
 REVISION_HEADER   = 1
 REVISION_ITEMS    = 2
 REVISION_BRANCHES = 3
 REVISION_COMMENT  = 4

 # These two functions will together split a list of items, separated
 # by sep, into a list, where each item has as much leading whitespace
 # as the whole original line had.
 def repeat(self, str, times):
  ret = ""
  for index in range(1, times):
   ret = ret + str
  return ret
 def splitNormalize(self, str, sep):
  items = split(str, sep)
  if len(items) > 1:
   leadingSpace = 0;
   while leadingSpace <= len(items[0]) and items[0][leadingSpace] in whitespace:
    leadingSpace = leadingSpace + 1
   for index in range(1, len(items)):
    items[index] = self.repeat(" ", leadingSpace) + strip(items[index])
  return items
 def __init__(self):
  PragmaticTree.__bases__[0].__init__(self)
  self.reading = PragmaticTree.HEADER
  self.revision = None
 def read(self, str):
  if str == self.fileSep or str == self.revisionSep:
   if self.revision:
    if not self.has_key(self.revisionsKey):
     self[self.revisionsKey] = []
    # Convert revision date to UNIX time. This is just for convinience
    self.revision[self.dateKey][0] = mktime(strptime(self.revision[self.dateKey][0], self.timeFormat))
    self[self.revisionsKey].append(self.revision)
   if str == self.fileSep:
    return TRUE
   else:
    self.revision = GeneralTree()
    self.reading = PragmaticTree.REVISION_HEADER
    return FALSE
  if self.reading == PragmaticTree.HEADER:
   for readLine in split(str, self.fieldSeparator):
    PragmaticTree.__bases__[0].read(self, readLine)
  elif self.reading == PragmaticTree.REVISION_HEADER:
   self.revision[self.revisionKey] = [split(str)[1], None]
   self.reading = PragmaticTree.REVISION_ITEMS
  elif self.reading == PragmaticTree.REVISION_ITEMS:
   for readLine in self.splitNormalize(str, self.fieldSeparator):
    self.revision.read(readLine)
   self.reading = PragmaticTree.REVISION_BRANCHES
  elif self.reading == PragmaticTree.REVISION_BRANCHES and find(str, self.branchesPrefix) != -1:
   for readLine in self.splitNormalize(str, self.fieldSeparator):
    self.revision.read(readLine)
   self.reading = PragmaticTree.REVISION_COMMENT
  else:
   if not self.revision.has_key(self.descriptionKey):
    self.revision[self.descriptionKey] = [str, None]
   else:
    self.revision[self.descriptionKey][0] = self.revision[self.descriptionKey][0] + "\n" + str
  return FALSE


class List:
 def __init__(self):
  self.files = []
  self.file = None
 def read(self, str):
  if not self.file:
   self.file = PragmaticTree()
  if self.file.read(str):
   self.files.append(self.file)
   self.file = None
 def __len__(self):
  return len(self.files)
 def __getitem__(self, key):
  return self.files[key]
 def __setitem__(self, key, value):
  self.files[key] = value
 def __delitem__(self, key):
  del self.files[key]
 def append(self, item):
  self.files.append(item)
 def __repr__(self):
  return repr(self.files)
